# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"BaseEvolver": "01_evolvers.ipynb",
         "EvolverWrapper": "01_evolvers.ipynb",
         "SigmoidEvolver": "01_evolvers.ipynb",
         "StandardEvolver": "01_evolvers.ipynb",
         "DynamicEvolver": "01_evolvers.ipynb",
         "check_hash_file_exists": "01_evolvers.ipynb",
         "evolve_from_properties_file": "01_evolvers.ipynb",
         "BaseConfiguration": "02_configs.ipynb",
         "ProportionalConfiguration": "02_configs.ipynb",
         "ConfigurationStructure": "02_configs.ipynb",
         "DynamicConfigurationStructure": "02_configs.ipynb",
         "DynamicConfiguration": "02_configs.ipynb",
         "BinaryOnes": "03_structure.ipynb",
         "BaseParameterType": "03_structure.ipynb",
         "Binary": "03_structure.ipynb",
         "Float": "03_structure.ipynb",
         "Literal": "03_structure.ipynb",
         "Smooth": "03_structure.ipynb",
         "ParameterFactory": "03_structure.ipynb",
         "StructureDefinition": "03_structure.ipynb",
         "generate_evolvers": "04_properties.ipynb",
         "generate_option_files": "04_properties.ipynb",
         "additional_properties": "04_properties.ipynb",
         "write_to_file": "04_properties.ipynb",
         "structure_parameters": "04_properties.ipynb",
         "configurable_parameters": "04_properties.ipynb",
         "description": "04_properties.ipynb",
         "fixed_parameters": "04_properties.ipynb"}

modules = ["examples.py",
           "evolvers.py",
           "configs.py",
           "structure.py",
           "properties.py"]

doc_url = "https://perceptualrobots.github.io/epct/"

git_url = "https://github.com/perceptualrobots/epct/tree/master/"

def custom_doc_links(name): return None
